(in-package #:lascivious)

(use-package "ALIEN")
(use-package "C-CALL")

(declaim (optimize (speed 3) (safety 0)))

;(def-alien-type "uchar*" (* unsigned-char))

(defstruct vox-coord
  (x 0 :type fixnum)
  (y 0 :type fixnum)
  (z 0 :type fixnum))

(defstruct pv
  (r 0   :type (unsigned-byte 8))
  (g 0   :type (unsigned-byte 8))
  (b 0   :type (unsigned-byte 8))
  (o 0.0 :type single-float))

(defstruct material
  (amb-r  0.25 :type single-float)
  (amb-g  0.25 :type single-float)
  (amb-b  0.25 :type single-float)
  (dif-r  0.25 :type single-float)
  (dif-g  0.25 :type single-float)
  (dif-b  0.25 :type single-float)
  (spec-r 0.25 :type single-float)
  (spec-g 0.25 :type single-float)
  (spec-b 0.25 :type single-float)
  (shinyness 10.0 :type single-float))

(defstruct lightsource
  (dir (normalize (make-vec3 :x 1.0d0 :y 1.0d0 :z -1.0d0)) :type vec3)
  (r 255 :type (unsigned-byte 8))
  (g 255 :type (unsigned-byte 8))
  (b 255 :type (unsigned-byte 8)))

(defconstant +max-lights+ 6)

(defstruct render-context
  (num-lights         1                                                    :type fixnum)
  (lights             (make-array +max-lights+ :element-type 'lightsource) :type simple-array)
  (projection         (mat4x4-identity)                                    :type mat4x4)

  )

(defvar *shade-funcs* (make-array (+ +max-lights+ 1) :element-type 'function))

(setf (aref *shade-funcs* 0) (build-shade-func 1))
(setf (aref *shade-funcs* 1) (build-shade-func 1))
(setf (aref *shade-funcs* 2) (build-shade-func 2))
(setf (aref *shade-funcs* 3) (build-shade-func 3))
(setf (aref *shade-funcs* 4) (build-shade-func 4))
(setf (aref *shade-funcs* 5) (build-shade-func 5))
(setf (aref *shade-funcs* 6) (build-shade-func 6))

(defun get-shade-func (num-lights)
  "Returns a shading function for the number of lights specified in NUM-LIGHTS"
  (aref *shade-funcs* num-lights))


(defconstant +num-scalar-values+ 256)
(defconstant +num-materials+ 6)
(defconstant +num-grad-values+ 222)

(defstruct volume
  (materials          (make-array-of-objects +num-materials+ (make-material))
		      :type simple-array)
  (scalar-opacity-lut (make-array +num-scalar-values+ :element-type 'single-float)
		      :type (simple-array single-float (*)))
  (grad-opacity-lut   (make-array +num-grad-values+ :element-type 'single-float)
		      :type (simple-array single-float (*)))
  (mat-lut            (make-array +num-scalar-values+ :element-type '(unsigned-byte 4))
		      :type (simple-array (unsigned-byte 4) (*)))

  ; currently unused
  (grad-lut           (make-array '(1 1 1) :element-type 'vec3)    :type simple-array)
  
  (vol-width      0                                            :type fixnum)
  (vol-height     0                                            :type fixnum)
  (vol-images     0                                            :type fixnum)  
  (vol-max-width  0.0d0                                        :type double-float)
  (vol-max-height 0.0d0                                        :type double-float)
  (vol-max-depth  0.0d0                                        :type double-float)
  
  (sampling-rate  2.0                                          :type single-float)
  (sr-opacity-adj 1.0                                          :type single-float)

  (interpolate-func    #'trilinear-interpolate :type (function (volume double-float double-float double-float)
							       fixnum))
  (calc-grad-func      #'calc-gradient         :type (function (volume vox-coord) vec3))
  (shade-func          (aref *shade-funcs* 1)  :type (function (render-context material vec3 single-float)
							       pv))
  
  (trans          (mat4x4-identity)                            :type mat4x4)
  (invtrans       (mat4x4-identity)                            :type mat4x4))


(defun set-material! (mat ar ag ab dr dg db sr sg sb shinyness)
  "Destructively modifies a material's values.
  MAT is the material to be modified.
  The rest of the args are the material's fields."
  (setf (material-amb-r mat) ar)
  (setf (material-amb-g mat) ag)
  (setf (material-amb-b mat) ab)
  (setf (material-dif-r mat) dr)
  (setf (material-dif-g mat) dg)
  (setf (material-dif-b mat) db)
  (setf (material-spec-r mat) sr)
  (setf (material-spec-g mat) sg)  
  (setf (material-spec-b mat) sb)
  (setf (material-shinyness mat) shinyness))









