(in-package #:lascivious)

(declaim (optimize (speed 3) (safety 0)))

(defmacro choose-one (func exp1 list)
  `())

(defun remove-empty-lists (lst)
  "Removes all occurrences of '() from a list"
  (cond ((null lst) nil)
	((null (car lst)) (remove-empty-lists (cdr lst)))
	(t (cons (car lst) (remove-empty-lists (cdr lst))))))
  
(defun ramp (start end points)
  "Returns a list of reals ramped from START to END of length POINTS"
  (labels ((ramp-helper (start end delta count)
			(cond ((= count 1) (cons end nil))
			      (t (cons start
				       (ramp-helper (+ start delta) end delta (- count 1)))))))
    (ramp-helper start end (/ (- end start) (- points 1)) points)))

(defmacro make-array-of-objects (size object)
  "Creates an array of seperate instances of OBJECT."
  `(let ((array (make-array ,size :element-type (type-of ,object))))
     (dotimes (i ,size)
       (setf (aref array i) ,object))
     array))

(defun bind-light-if-exists (light numlights)
  (if (> numlights light)
      (cond ((= light 0) `(light0 (aref (render-context-lights context) ,light)))
	    ((= light 1) `(light1 (aref (render-context-lights context) ,light)))
	    ((= light 2) `(light2 (aref (render-context-lights context) ,light)))
	    ((= light 3) `(light3 (aref (render-context-lights context) ,light)))
	    ((= light 4) `(light4 (aref (render-context-lights context) ,light)))
	    ((= light 5) `(light5 (aref (render-context-lights context) ,light)))
	    (t nil))
    nil))

(defun bind-dotprod-if-light-exists (light numlights)
  (if (> numlights light)
      (cond ((= light 0) `(dotprod0 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light)))))
	    ((= light 1) `(dotprod1 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light)))))
	    ((= light 2) `(dotprod2 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light)))))
	    ((= light 3) `(dotprod3 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light)))))
	    ((= light 4) `(dotprod4 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light)))))
	    ((= light 5) `(dotprod5 (dot gradient
					 (lightsource-dir (aref (render-context-lights context)
								,light))))))
    nil))

(defun get-light-var (num)
  (cond ((= num 0) 'light0)
	((= num 1) 'light1)
	((= num 2) 'light2)
	((= num 3) 'light3)
	((= num 4) 'light4)
	((= num 5) 'light5)))

(defun get-dotprod-var (num)
  (cond ((= num 0) 'dotprod0)
	((= num 1) 'dotprod1)
	((= num 2) 'dotprod2)
	((= num 3) 'dotprod3)
	((= num 4) 'dotprod4)
	((= num 5) 'dotprod5)))

(defun get-shading-code (term num-lights) 
  (cond ((equal term "red") `(+ ,@(do ((i 0 (+ i 1))
				       (code '()))
				      ((= i num-lights) code)
				    (setf code (append code
						       `((the fixnum (truncate (* (material-amb-r mat)
									       (lightsource-r ,(get-light-var i)))))
							 (the fixnum (truncate (* (material-dif-r mat)
										  (lightsource-r ,(get-light-var i))
										  ,(get-dotprod-var i))))))))))
	((equal term "green") `(+ ,@(do ((i 0 (+ i 1))
					 (code '()))
					((= i num-lights) code)
				      (setf code (append code
							 `((the fixnum (truncate (* (material-amb-g mat)
										    (lightsource-g ,(get-light-var i)))))
							   (the fixnum (truncate (* (material-dif-g mat)
										    (lightsource-g ,(get-light-var i))
										    ,(get-dotprod-var i))))))))))
	((equal term "blue") `(+ ,@(do ((i 0 (+ i 1))
					(code '()))
				       ((= i num-lights) code)
				     (setf code (append code
							`((the fixnum (truncate (* (material-amb-b mat)
										   (lightsource-b ,(get-light-var i)))))
							  (the fixnum (truncate (* (material-dif-b mat)
										   (lightsource-b ,(get-light-var i))
										   ,(get-dotprod-var i))))))))))))

(defmacro build-shade-func (num-lights)
  "Builds a shading function with NUM-LIGHTS light sources."
  `(lambda (context mat gradient opacity)
     (let (,@(remove-empty-lists (mapcar #'bind-light-if-exists
					 '(0 1 2 3 4 5)
					 `(,num-lights ,num-lights ,num-lights
					   ,num-lights ,num-lights ,num-lights)))
	   ,@(remove-empty-lists (mapcar #'bind-dotprod-if-light-exists
					 '(0 1 2 3 4 5)
					 `(,num-lights ,num-lights ,num-lights
					   ,num-lights ,num-lights ,num-lights))))
       ,@(funcall #'(lambda (num-lights)
		      (let ((code nil))
			(dotimes (i num-lights)
			  (setf code (append `,code `((if (< ,(get-dotprod-var i) 0.0d0)
							  (setf ,(get-dotprod-var i) 0.0d0))))))
			code))
		  num-lights)
       (make-pv :r ,(get-shading-code "red" num-lights)
		:g ,(get-shading-code "green" num-lights)
		:b ,(get-shading-code "blue" num-lights)
		:o opacity))))
       

(defmacro set-opacity-lut (table &rest lst)
  (let ((len (length lst)))
    (pprint lst)
    (format t "~%len = ~A~%" len)
    (cond ((= len 3) `(let ((lut ,table))
			(setf (aref lut (first '(,@lst))) (third '(,@lst)))))
	  ((= len 5) `(let ((lut ,table))
			(do ((i (first '(,@lst)) (+ i 1)))
			    ((> i (third '(,@lst))))
			  (setf (aref lut i) (fifth '(,@lst))))))
	  ((= len 7) `(let ((r (ramp (fifth '(,@lst))
				     (seventh '(,@lst))
				     (+ (- (third '(,@lst)) (first '(,@lst)))
					1)))
			    (lut ,table))
			(do ((i (first '(,@lst)) (+ i 1)))
			    ((> i (third '(,@lst))))
			  (setf (aref lut i) (pop r)))))
	  (t nil))))

(defmacro set-material-lut (table &rest lst)
  (let ((len (length lst)))
    (pprint lst)
    (format t "~%len = ~A~%" len)
    (cond ((= len 3) `(let ((lut ,table))
			(setf (aref lut (first '(,@lst))) (third '(,@lst)))))
	  ((= len 5) `(let ((lut ,table))
			(do ((i (first '(,@lst)) (+ i 1)))
			    ((> i (third '(,@lst))))
			  (setf (aref lut i) (fifth '(,@lst))))))
	  (t nil))))



;;; Optimization macros

;; just declare a simple way to make a global integer
(defmacro defint(name &optional ( value 0))
  `(progn (defparameter ,name ,value)
	  (declaim (fixnum ,name))))

;; this defines a fast function.  It does several things.
;; A fast function can be declared to be inline, and it can
;; have a type.  Its parameters can be typed.
(defmacro deffun( name params &rest body )
  (unless (consp name) (setf name (list name)))
  (let ((type 'atom))
    (when (= (length name) 2)
      (setf type (pop name)))
    `(progn
       (defun ,(first name)
	 ,@(let ((names (mapcar (lambda(x)
				  (if (consp x) (first x) x))
				params)))
	     `(,names ,@(loop for p in params when
			      (consp p) collect `(declare (,(second p) ,(first p))))))
	 (declare (optimize (speed 3)(safety 0)(fixnum-safety 0)))
	 (the ,type (progn ,@body))))))

(defmacro deffuni( name &rest rst )
  `(progn
     (declaim (inline ,(if (consp name)
			   (first (reverse name))
			 name)))
     (deffun ,name ,@rst)))

;; a way to do a funcall that doesn't check to see if the target may be a symbol
(defmacro ffuncall(target &rest args)
  `(funcall (the function ,target) ,@args))


